/*
    Qalculate    

    Copyright (C) 2003  Niklas Knutsson (nq@altern.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>

#include "data/icon.xpm"

#include <libqalculate/qalculate.h>
#include <support.h>

GdkPixbuf *icon_pixbuf;

bool first_time, first_qalculate_run;
string wget_args;
string initial_from_unit, initial_to_unit;
int initial_width, initial_height;
bool initially_expanded;

GtkWidget *units_dialog;
GtkWidget *units_treeview_category;
GtkWidget *units_treeview_unit1;
GtkWidget *units_treeview_unit2;
GtkWidget *units_label_equals;
GtkWidget *units_button_convert;
GtkWidget *units_button_reverse;
GtkWidget *units_entry_to_val;
GtkWidget *units_entry_from_val;
GtkWidget *units_button_close;
GtkWidget *units_entry_from_unit;
GtkWidget *units_entry_to_unit;
GtkWidget *units_expander;

GtkTreeStore *categories_store;
GtkListStore *units_store;
string selected_unit_category;

bool block_unit_convert;

#define EXPAND_TO_ITER(model, view, iter)		GtkTreePath *path = gtk_tree_model_get_path(model, &iter); \
							gtk_tree_view_expand_to_path(GTK_TREE_VIEW(view), path); \
							gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(view), path, NULL, FALSE, 0.0, 0.0);\
							gtk_tree_path_free(path);
#define EXPAND_ITER(model, view, iter)			GtkTreePath *path = gtk_tree_model_get_path(model, &iter); \
							gtk_tree_view_expand_row(GTK_TREE_VIEW(view), path, false); \
							gtk_tree_path_free(path);

enum {
	UNITS_TITLE_COLUMN,
	UNITS_NAMES_COLUMN,
	UNITS_POINTER_COLUMN,
	UNITS_N_COLUMNS
};

struct tree_struct {
	string item;
	list<tree_struct> items;
	list<tree_struct>::iterator it;
	list<tree_struct>::reverse_iterator rit;
	vector<void*> objects;	
	tree_struct *parent;
	void sort() {
		items.sort();
		for(list<tree_struct>::iterator it = items.begin(); it != items.end(); ++it) {
			it->sort();
		}
	}
	bool operator < (tree_struct &s1) const {
		return item < s1.item;	
	}	
};

tree_struct unit_cats;

EvaluationOptions evalops;
PrintOptions printops;

gint string_sort_func(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer user_data) {
	gint cid = GPOINTER_TO_INT(user_data);
	gchar *gstr1, *gstr2;
	gint retval;
	gtk_tree_model_get(model, a, cid, &gstr1, -1);
	gtk_tree_model_get(model, b, cid, &gstr2, -1);
	retval = g_ascii_strcasecmp(gstr1, gstr2);
	g_free(gstr1);
	g_free(gstr2);
	return retval;
}

PangoCoverageLevel get_least_coverage(const gchar *gstr, PangoCoverage *coverage) {

	if(!coverage) return PANGO_COVERAGE_EXACT;
	PangoCoverageLevel level = PANGO_COVERAGE_EXACT;
	while(gstr[0] != '\0') {
		if(gstr[0] < 0) {
			gunichar gu = g_utf8_get_char_validated(gstr, -1);
			if(gu != (gunichar) -1 && gu != (gunichar) -2) {
				if(pango_coverage_get(coverage, (int) gu) < level) {
					level = pango_coverage_get(coverage, gu);
				}
			}
		}
		gstr = g_utf8_find_next_char(gstr, NULL);
		if(!gstr) break;
	}
	return level;

}

PangoCoverageLevel get_least_coverage(const gchar *gstr, GtkWidget *widget) {
	
	PangoContext *context = gtk_widget_get_pango_context(widget);
	PangoFont *font = pango_context_load_font(context, gtk_widget_get_style(widget)->font_desc);
	PangoLanguage *language = pango_context_get_language(context);
	return get_least_coverage(gstr, pango_font_get_coverage(font, language));

}

bool can_display_unicode_string_function(const char *str, void *w) {
	if(!w) w = (void*) units_dialog;
	return get_least_coverage(str, (GtkWidget*) w) >= PANGO_COVERAGE_APPROXIMATE;
}

/*
	display errors generated under calculation
*/
void display_errors() {
	if(!CALCULATOR->message()) return;
	bool b = false, critical = false;
	MessageType mtype;
	GtkWidget *edialog;
	string str = "";
	while(true) {
		mtype = CALCULATOR->message()->type();
		if(mtype == MESSAGE_ERROR || mtype == MESSAGE_WARNING) {
			if(b) {
				str += "\n";
			} else {
				b = true;
			}
			str += CALCULATOR->message()->message();
			if(mtype == MESSAGE_ERROR) {
				critical = true;
			}
		} else {
			edialog = gtk_message_dialog_new(
					GTK_WINDOW(units_dialog),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_INFO,
					GTK_BUTTONS_CLOSE,
					CALCULATOR->message()->message().c_str());
			gtk_dialog_run(GTK_DIALOG(edialog));
			gtk_widget_destroy(edialog);
		}
		if(!CALCULATOR->nextMessage()) break;
	}
	if(!str.empty()) {
		if(critical) {
			edialog = gtk_message_dialog_new(
					GTK_WINDOW(units_dialog),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_CLOSE,
					str.c_str());
		} else {
			edialog = gtk_message_dialog_new(
					GTK_WINDOW(units_dialog),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_WARNING,
					GTK_BUTTONS_CLOSE,
					str.c_str());
		}

		gtk_dialog_run(GTK_DIALOG(edialog));
		gtk_widget_destroy(edialog);
	}
}

gboolean on_display_errors_timeout(gpointer) {
	display_errors();
	return true;
}


void load_preferences();
void save_preferences();
void update_units_tree();
void generate_units_tree_struct();


#ifdef __cplusplus
extern "C" {
#endif

void convert_units(int toFrom = -1);
void on_units_treeview_category_selection_changed(GtkTreeSelection *treeselection, gpointer user_data);
void on_units_treeview_unit1_selection_changed(GtkTreeSelection *treeselection, gpointer user_data);
void on_units_treeview_unit2_selection_changed(GtkTreeSelection *treeselection, gpointer user_data);
void on_units_button_close_clicked(GtkButton *button, gpointer user_data);
void close_app();
void on_units_button_convert_clicked(GtkButton *button, gpointer user_data);
void on_units_button_reverse_clicked(GtkButton *button, gpointer user_data);
void on_units_entry_from_val_activate(GtkEntry *entry, gpointer user_data);
void on_units_entry_to_val_activate(GtkEntry *entry, gpointer user_data);
void on_units_entry_from_val_changed(GtkEditable *entry, gpointer user_data);
void on_units_entry_to_val_changed(GtkEditable *entry, gpointer user_data);
void on_units_entry_from_unit_activate(GtkEntry *entry, gpointer user_data);
void on_units_entry_to_unit_activate(GtkEntry *entry, gpointer user_data);
void on_units_entry_from_unit_changed(GtkEditable *entry, gpointer user_data);
void on_units_entry_to_unit_changed(GtkEditable *entry, gpointer user_data);

void on_units_button_close_clicked(GtkButton*, gpointer) {
	close_app();
}

void close_app() {
	save_preferences();
	gtk_main_quit();
}

void on_units_button_convert_clicked(GtkButton*, gpointer) {
	convert_units(0);
	gtk_widget_grab_focus(units_entry_from_val);
}
void on_units_button_reverse_clicked(GtkButton*, gpointer) {
	convert_units(1);
	gtk_widget_grab_focus(units_entry_to_val);
}
void on_units_entry_from_val_activate(GtkEntry*, gpointer) {
	convert_units(0);
}
void on_units_entry_to_val_activate(GtkEntry*, gpointer) {
	convert_units(1);
}
void on_units_entry_from_val_changed(GtkEditable*, gpointer) {
	g_signal_handlers_block_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
	gtk_entry_set_text(GTK_ENTRY(units_entry_to_val), "");
	g_signal_handlers_unblock_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
}
void on_units_entry_to_val_changed(GtkEditable*, gpointer) {
	g_signal_handlers_block_matched((gpointer) units_entry_from_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_from_val_changed, NULL);
	gtk_entry_set_text(GTK_ENTRY(units_entry_from_val), "");
	g_signal_handlers_unblock_matched((gpointer) units_entry_from_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_from_val_changed, NULL);
}
void on_units_entry_from_unit_activate(GtkEntry*, gpointer) {
	convert_units(0);
}
void on_units_entry_to_unit_activate(GtkEntry*, gpointer) {
	convert_units(0);
}
void on_units_entry_from_unit_changed(GtkEditable*, gpointer) {
	g_signal_handlers_block_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
	gtk_entry_set_text(GTK_ENTRY(units_entry_to_val), "");
	g_signal_handlers_unblock_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
}
void on_units_entry_to_unit_changed(GtkEditable*, gpointer) {
	g_signal_handlers_block_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
	gtk_entry_set_text(GTK_ENTRY(units_entry_to_val), "");
	g_signal_handlers_unblock_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
}

void convert_units(int toFrom) {

	string str1 = gtk_entry_get_text(GTK_ENTRY(units_entry_from_unit));
	string str2 = gtk_entry_get_text(GTK_ENTRY(units_entry_to_unit));
	remove_blank_ends(str1);	
	remove_blank_ends(str2);
	
	CompositeUnit *cu1 = NULL;
	Unit *u1 = NULL;
	if(!str1.empty()) {
		u1 = CALCULATOR->getUnit(str1);
		if(!u1) {
			cu1 = new CompositeUnit("", "temporary_composite_convert1", "", str1);
			if(cu1->get(1)) {
				u1 = cu1;
			} else {
				delete cu1;
				gtk_widget_grab_focus(units_entry_from_unit);
				return;
			}
		}
	}
	
	CompositeUnit *cu2 = NULL;
	Unit *u2 = NULL;
	if(!str2.empty()) {
		u2 = CALCULATOR->getUnit(str2);
		if(!u2) {
			cu2 = new CompositeUnit("", "temporary_composite_convert2", "", str2);
			if(cu2->get(1)) {
				u2 = cu2;
			} else {
				delete cu2;
				gtk_widget_grab_focus(units_entry_to_unit);
				return;
			}
		}
	}
	
	//values
	const gchar *fromValue = gtk_entry_get_text(GTK_ENTRY(units_entry_from_val));
	const gchar *toValue = gtk_entry_get_text(GTK_ENTRY(units_entry_to_val));
	//determine conversion direction
	bool b = false;	
	printops.is_approximate = &b;
	g_signal_handlers_block_matched((gpointer) units_entry_from_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_from_val_changed, NULL);
	g_signal_handlers_block_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
	if(toFrom > 0) {
		MathStructure v_mstruct;
		CALCULATOR->parse(&v_mstruct, CALCULATOR->unlocalizeExpression(toValue, evalops.parse_options), evalops.parse_options);
		if(u2) {
			v_mstruct *= u2;
		}
		v_mstruct.eval(evalops);
		if(u1) {
			v_mstruct.convert(u1, true);
			v_mstruct.divide(u1, true);
			v_mstruct.eval(evalops);
		}
		printops.can_display_unicode_string_arg = (void*) units_entry_from_val;
		v_mstruct.format(printops);
		gtk_entry_set_text(GTK_ENTRY(units_entry_from_val), v_mstruct.print(printops).c_str());
		b = b || v_mstruct.isApproximate();
	} else {
		MathStructure v_mstruct;
		CALCULATOR->parse(&v_mstruct, CALCULATOR->unlocalizeExpression(fromValue, evalops.parse_options), evalops.parse_options);
		if(u1) {
			v_mstruct *= u1;
		}
		v_mstruct.eval(evalops);
		if(u2) {
			v_mstruct.convert(u2, true);
			v_mstruct.divide(u2, true);
			v_mstruct.eval(evalops);
		}
		printops.can_display_unicode_string_arg = (void*) units_entry_to_val;
		v_mstruct.format(printops);
		gtk_entry_set_text(GTK_ENTRY(units_entry_to_val), v_mstruct.print(printops).c_str());
		b = b || v_mstruct.isApproximate();
	}
	
	if(b && can_display_unicode_string_function(SIGN_ALMOST_EQUAL, (void*) units_label_equals)) {
		gtk_label_set_text(GTK_LABEL(units_label_equals), SIGN_ALMOST_EQUAL);
	} else {
		gtk_label_set_text(GTK_LABEL(units_label_equals), "=");
	}
		
	if(cu1) delete cu1;
	if(cu2) delete cu2;
	
	g_signal_handlers_unblock_matched((gpointer) units_entry_from_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_from_val_changed, NULL);
	g_signal_handlers_unblock_matched((gpointer) units_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_entry_to_val_changed, NULL);
	
}



#ifdef __cplusplus
}
#endif


void generate_units_tree_struct() {
	size_t cat_i, cat_i_prev; 
	bool b;	
	string str, cat, cat_sub;
	Unit *u = NULL;
	unit_cats.items.clear();
	unit_cats.objects.clear();
	unit_cats.parent = NULL;	
	list<tree_struct>::iterator it;	
	for(size_t i = 0; i < CALCULATOR->units.size(); i++) {
		if(CALCULATOR->units[i]->isActive() && !CALCULATOR->units[i]->isHidden() && CALCULATOR->units[i] != CALCULATOR->u_euro) {
			tree_struct *item = &unit_cats;
			if(!CALCULATOR->units[i]->category().empty()) {
				cat = CALCULATOR->units[i]->category();
				cat_i = cat.find("/"); cat_i_prev = 0;
				b = false;
				while(true) {
					if(cat_i == string::npos) {
						cat_sub = cat.substr(cat_i_prev, cat.length() - cat_i_prev);
					} else {
						cat_sub = cat.substr(cat_i_prev, cat_i - cat_i_prev);
					}
					b = false;
					for(it = item->items.begin(); it != item->items.end(); ++it) {
						if(cat_sub == it->item) {
							item = &*it;
							b = true;
							break;
						}
					}
					if(!b) {
						tree_struct cat;		
						item->items.push_back(cat);
						it = item->items.end();
						--it;
						it->parent = item;
						item = &*it;
						item->item = cat_sub;
					}
					if(cat_i == string::npos) {
						break;
					}
					cat_i_prev = cat_i + 1;
					cat_i = cat.find("/", cat_i_prev);
				}
			}
			b = false;
			for(size_t i3 = 0; i3 < item->objects.size(); i3++) {
				u = (Unit*) item->objects[i3];
				if(CALCULATOR->units[i]->title() < u->title()) {
					b = true;
					item->objects.insert(item->objects.begin() + i3, (void*) CALCULATOR->units[i]);
					break;
				}
			}
			if(!b) item->objects.push_back((void*) CALCULATOR->units[i]);		
		}
	}
	
	unit_cats.sort();

}

void update_units_tree() {

	GtkTreeIter iter, iter2, iter3;
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(units_treeview_category));
	GtkTreeSelection *select = gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_category));
	g_signal_handlers_block_matched((gpointer) select, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_treeview_category_selection_changed, NULL);
	gtk_tree_store_clear(categories_store);
	g_signal_handlers_unblock_matched((gpointer) select, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_treeview_category_selection_changed, NULL);
	gtk_tree_store_append(categories_store, &iter3, NULL);
	gtk_tree_store_set(categories_store, &iter3, 0, _("All"), 1, _("All"), -1);
	string str;
	tree_struct *item, *item2;
	unit_cats.it = unit_cats.items.begin();
	if(unit_cats.it != unit_cats.items.end()) {
		item = &*unit_cats.it;
		++unit_cats.it;
		item->it = item->items.begin();
	} else {
		item = NULL;
	}
	str = "";
	iter2 = iter3;
	while(item) {
		gtk_tree_store_append(categories_store, &iter, &iter2);
		str += "/";
		str += item->item;
		gtk_tree_store_set(categories_store, &iter, 0, item->item.c_str(), 1, str.c_str(), -1);
		if(str == selected_unit_category) {
			EXPAND_TO_ITER(model, units_treeview_category, iter)
			gtk_tree_selection_select_iter(gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_category)), &iter);
		}
		while(item && item->it == item->items.end()) {
			size_t str_i = str.rfind("/");
			if(str_i == string::npos) {
				str = "";
			} else {
				str = str.substr(0, str_i);
			}
			item = item->parent;
			gtk_tree_model_iter_parent(model, &iter2, &iter);
			iter = iter2;
		}
		if(item) {
			item2 = &*item->it;
			if(item->it == item->items.begin()) iter2 = iter;
			++item->it;
			item = item2;
			item->it = item->items.begin();	
		}
	}
	if(!unit_cats.objects.empty()) {	
		//add "Uncategorized" category if there are units without category
		gtk_tree_store_append(categories_store, &iter, &iter3);
		gtk_tree_store_set(categories_store, &iter, 0, _("Uncategorized"), 1, _("Uncategorized"), -1);
		if(selected_unit_category == _("Uncategorized")) {
			EXPAND_TO_ITER(model, units_treeview_category, iter)
			gtk_tree_selection_select_iter(gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_category)), &iter);
		}
	}	
	if(!gtk_tree_selection_get_selected(gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_category)), &model, &iter)) {
		//if no category has been selected (previously selected has been renamed/deleted), select "All"
		selected_unit_category = _("All");
		gtk_tree_model_get_iter_first(GTK_TREE_MODEL(categories_store), &iter);
		EXPAND_ITER(model, units_treeview_category, iter)
		gtk_tree_selection_select_iter(gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_category)), &iter);
	}
}

void setUnitTreeItem(GtkTreeIter &iter2, Unit *u) {
	gtk_list_store_append(units_store, &iter2);
	string snames;
	//display name, plural name and short name in the second column
	switch(u->subtype()) {
		case SUBTYPE_COMPOSITE_UNIT: {
			snames = ((CompositeUnit*) u)->print(false, true, true, &can_display_unicode_string_function, (void*) units_treeview_unit1);
			snames += ", ";
			snames += ((CompositeUnit*) u)->print(false, false, true, &can_display_unicode_string_function, (void*) units_treeview_unit1);
			break;
		}
		case SUBTYPE_ALIAS_UNIT: {}
		case SUBTYPE_BASE_UNIT: {
			for(size_t i = 1; i <= u->countNames(); i++) {
				if(i > 1) snames += ", ";
				snames += u->getName(i).name;
			}
			break;
		}
	}
	//display descriptive name (title), or name if no title defined
	gtk_list_store_set(units_store, &iter2, UNITS_TITLE_COLUMN, u->title(true).c_str(), UNITS_NAMES_COLUMN, snames.c_str(), UNITS_POINTER_COLUMN, (gpointer) u, -1);
}


void on_units_treeview_category_selection_changed(GtkTreeSelection *treeselection, gpointer) {
	GtkTreeModel *model;
	GtkTreeIter iter, iter2;
	//make sure that no unit conversion is done in the dialog until everthing is updated
	block_unit_convert = true;

	bool no_cat = false, b_all = false;
	GtkTreeSelection *select1 = gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_unit1));
	g_signal_handlers_block_matched((gpointer) select1, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_treeview_unit1_selection_changed, NULL);
	GtkTreeSelection *select2 = gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_unit2));
	g_signal_handlers_block_matched((gpointer) select2, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_treeview_unit2_selection_changed, NULL);
	gtk_list_store_clear(units_store);
	g_signal_handlers_unblock_matched((gpointer) select1, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_treeview_unit1_selection_changed, NULL);
	g_signal_handlers_unblock_matched((gpointer) select2, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_units_treeview_unit2_selection_changed, NULL);
	if(gtk_tree_selection_get_selected(treeselection, &model, &iter)) {
		gchar *gstr;
		gtk_tree_model_get(model, &iter, 1, &gstr, -1);
		selected_unit_category = gstr;
		if(selected_unit_category == _("All")) {
			b_all = true;
		} else if(selected_unit_category == _("Uncategorized")) {
			no_cat = true;
		}
		if(!b_all && !no_cat && selected_unit_category[0] == '/') {
			string str = selected_unit_category.substr(1, selected_unit_category.length() - 1);
			for(size_t i = 0; i < CALCULATOR->units.size(); i++) {	
				if(CALCULATOR->units[i]->isActive() && !CALCULATOR->units[i]->isHidden() && CALCULATOR->units[i]->category().substr(0, selected_unit_category.length() - 1) == str) {
					setUnitTreeItem(iter2, CALCULATOR->units[i]);
				}
			}
		} else {
			for(size_t i = 0; i < CALCULATOR->units.size(); i++) {
				if(CALCULATOR->units[i]->isActive() && !CALCULATOR->units[i]->isHidden() && (b_all || (no_cat && CALCULATOR->units[i]->category().empty()) || CALCULATOR->units[i]->category() == selected_unit_category)) {
					setUnitTreeItem(iter2, CALCULATOR->units[i]);
				}
			}
		}
		g_free(gstr);
	} else {
		selected_unit_category = "";
	}
	block_unit_convert = false;
	convert_units();
}

void on_units_treeview_unit1_selection_changed(GtkTreeSelection *treeselection, gpointer) {
	GtkTreeModel *model;
	GtkTreeIter iter;
	if(gtk_tree_selection_get_selected(treeselection, &model, &iter)) {
		Unit *u;
		gtk_tree_model_get(model, &iter, UNITS_POINTER_COLUMN, &u, -1);
		gtk_entry_set_text(GTK_ENTRY(units_entry_from_unit), u->print(true, true, true, &can_display_unicode_string_function, (void*) units_entry_from_unit).c_str());
		if(!block_unit_convert) convert_units();
		gtk_tree_selection_unselect_all(treeselection);
		gtk_widget_grab_focus(units_entry_from_val);
	}
}
void on_units_treeview_unit2_selection_changed(GtkTreeSelection *treeselection, gpointer) {
	GtkTreeModel *model;
	GtkTreeIter iter;
	if(gtk_tree_selection_get_selected(treeselection, &model, &iter)) {
		Unit *u;
		gtk_tree_model_get(model, &iter, UNITS_POINTER_COLUMN, &u, -1);
		gtk_entry_set_text(GTK_ENTRY(units_entry_to_unit), u->print(true, true, true, &can_display_unicode_string_function, (void*) units_entry_to_unit).c_str());
		if(!block_unit_convert) convert_units();
		gtk_tree_selection_unselect_all(treeselection);
		gtk_widget_grab_focus(units_entry_from_val);
	}
}


GtkWidget *create_units_dialog(void) {

	GtkWidget *dialog_vbox3;
	GtkWidget *hbox101;
	GtkWidget *vbox59;
	GtkWidget *table7;
	GtkWidget *label93;
	GtkWidget *label94;
	GtkWidget *vbox60;
	GtkWidget *frame3;
	GtkWidget *units_scrolledwindow1;
	GtkWidget *label82;
	GtkWidget *hbox103;
	GtkWidget *frame2;
	GtkWidget *units_scrolledwindow2;
	GtkWidget *label81;
	GtkWidget *frame11;
	GtkWidget *scrolledwindow1;
	GtkWidget *label92;
	GtkWidget *label90;
	GtkWidget *dialog_action_area3;
	GtkWidget *alignment17;
	GtkWidget *hbox104;
	GtkWidget *image38;
	GtkWidget *label96;
	GtkWidget *alignment18;
	GtkWidget *hbox105;
	GtkWidget *image39;
	GtkWidget *label97;
	GtkTooltips *tooltips;

	tooltips = gtk_tooltips_new ();

	units_dialog = gtk_dialog_new ();
	gtk_container_set_border_width (GTK_CONTAINER (units_dialog), 5);
	gtk_window_set_title (GTK_WINDOW (units_dialog), _("Unit Conversion"));
	//gtk_window_set_default_size (GTK_WINDOW (units_dialog), initial_width, initial_height);
	gtk_window_set_resizable (GTK_WINDOW (units_dialog), FALSE);
	gtk_window_set_type_hint (GTK_WINDOW (units_dialog), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_dialog_set_has_separator (GTK_DIALOG (units_dialog), FALSE);

	dialog_vbox3 = GTK_DIALOG (units_dialog)->vbox;
	gtk_widget_show (dialog_vbox3);

	hbox101 = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (hbox101);
	gtk_box_pack_start (GTK_BOX (dialog_vbox3), hbox101, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (hbox101), 6);

	vbox59 = gtk_vbox_new (FALSE, 6);
	gtk_widget_show (vbox59);
	gtk_box_pack_start (GTK_BOX (hbox101), vbox59, TRUE, TRUE, 0);

	table7 = gtk_table_new (3, 3, FALSE);
	gtk_widget_show (table7);
	gtk_box_pack_start (GTK_BOX (vbox59), table7, FALSE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (table7), 6);
	gtk_table_set_row_spacings (GTK_TABLE (table7), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table7), 6);

	units_entry_to_val = gtk_entry_new();
	gtk_entry_set_alignment(GTK_ENTRY (units_entry_to_val), 1.0);
	gtk_widget_show (units_entry_to_val);
	gtk_table_attach (GTK_TABLE (table7), units_entry_to_val, 1, 2, 2, 3, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, units_entry_to_val, _("Converted value"), NULL);
	gtk_entry_set_width_chars (GTK_ENTRY (units_entry_to_val), 30);

	units_entry_from_val = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(units_entry_from_val), "1");
	gtk_entry_set_alignment(GTK_ENTRY (units_entry_from_val), 1.0);
	gtk_widget_show (units_entry_from_val);
	gtk_table_attach (GTK_TABLE (table7), units_entry_from_val, 1, 2, 1, 2, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, units_entry_from_val, _("Value to convert from"), NULL);
	gtk_entry_set_width_chars (GTK_ENTRY (units_entry_from_val), 30);

	label93 = gtk_label_new (_("Value"));
	gtk_widget_show (label93);
	gtk_table_attach (GTK_TABLE (table7), label93, 1, 2, 0, 1, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label93), 0, 0.5);

	label94 = gtk_label_new (_("Unit"));
	gtk_widget_show (label94);
	gtk_table_attach (GTK_TABLE (table7), label94, 2, 3, 0, 1, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label94), 0, 0.5);

	units_entry_from_unit = gtk_entry_new ();
	gtk_entry_set_text(GTK_ENTRY(units_entry_from_unit), initial_from_unit.c_str());
	gtk_entry_set_alignment(GTK_ENTRY (units_entry_from_unit), 0.0);
	gtk_widget_show (units_entry_from_unit);
	gtk_table_attach (GTK_TABLE (table7), units_entry_from_unit, 2, 3, 1, 2, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, units_entry_from_unit, _("Unit expression to convert from"), NULL);
	gtk_entry_set_width_chars (GTK_ENTRY (units_entry_from_unit), 20);

	units_entry_to_unit = gtk_entry_new ();
	gtk_entry_set_text(GTK_ENTRY(units_entry_to_unit), initial_to_unit.c_str());
	gtk_entry_set_alignment(GTK_ENTRY (units_entry_to_unit), 0.0);
	gtk_widget_show (units_entry_to_unit);
	gtk_table_attach (GTK_TABLE (table7), units_entry_to_unit, 2, 3, 2, 3, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, units_entry_to_unit, _("Unit expression to convert to"), NULL);
	gtk_entry_set_width_chars (GTK_ENTRY (units_entry_to_unit), 20);

	units_label_equals = gtk_label_new (_("="));
	gtk_widget_show (units_label_equals);
	gtk_table_attach (GTK_TABLE (table7), units_label_equals, 0, 1, 2, 3, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_label_set_justify (GTK_LABEL (units_label_equals), GTK_JUSTIFY_RIGHT);
	gtk_misc_set_alignment (GTK_MISC (units_label_equals), 1, 0.5);

	units_expander = gtk_expander_new (NULL);
	gtk_widget_show (units_expander);
	gtk_box_pack_start (GTK_BOX (vbox59), units_expander, TRUE, TRUE, 0);
	gtk_expander_set_expanded (GTK_EXPANDER (units_expander), initially_expanded);
	gtk_expander_set_spacing (GTK_EXPANDER (units_expander), 6);

	vbox60 = gtk_vbox_new (FALSE, 6);
	gtk_widget_show (vbox60);
	gtk_container_add (GTK_CONTAINER (units_expander), vbox60);

	frame3 = gtk_frame_new (NULL);
	gtk_widget_show (frame3);
	gtk_box_pack_start (GTK_BOX (vbox60), frame3, TRUE, TRUE, 0);

	units_scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (units_scrolledwindow1);
	gtk_container_add (GTK_CONTAINER (frame3), units_scrolledwindow1);
	gtk_container_set_border_width (GTK_CONTAINER (units_scrolledwindow1), 6);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (units_scrolledwindow1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (units_scrolledwindow1), GTK_SHADOW_ETCHED_IN);

	units_treeview_category = gtk_tree_view_new ();
	gtk_widget_show (units_treeview_category);
	gtk_container_add (GTK_CONTAINER (units_scrolledwindow1), units_treeview_category);
	gtk_widget_set_size_request (units_treeview_category, -1, 150);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (units_treeview_category), FALSE);
	
	label82 = gtk_label_new (_("Category"));
	gtk_widget_show (label82);
	gtk_frame_set_label_widget (GTK_FRAME (frame3), label82);

	hbox103 = gtk_hbox_new (FALSE, 6);
	gtk_widget_show (hbox103);
	gtk_box_pack_start (GTK_BOX (vbox60), hbox103, TRUE, TRUE, 0);

	frame2 = gtk_frame_new (NULL);
	gtk_widget_show (frame2);
	gtk_box_pack_start (GTK_BOX (hbox103), frame2, TRUE, TRUE, 0);

	units_scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (units_scrolledwindow2);
	gtk_container_add (GTK_CONTAINER (frame2), units_scrolledwindow2);
	gtk_container_set_border_width (GTK_CONTAINER (units_scrolledwindow2), 6);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (units_scrolledwindow2), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (units_scrolledwindow2), GTK_SHADOW_ETCHED_IN);

	units_treeview_unit1 = gtk_tree_view_new ();
	gtk_widget_show (units_treeview_unit1);
	gtk_container_add (GTK_CONTAINER (units_scrolledwindow2), units_treeview_unit1);
	gtk_widget_set_size_request (units_treeview_unit1, 230, 200);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (units_treeview_unit1), FALSE);

	label81 = gtk_label_new (_("From Unit"));
	gtk_widget_show (label81);
	gtk_frame_set_label_widget (GTK_FRAME (frame2), label81);

	frame11 = gtk_frame_new (NULL);
	gtk_widget_show (frame11);
	gtk_box_pack_start (GTK_BOX (hbox103), frame11, TRUE, TRUE, 0);

	scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (scrolledwindow1);
	gtk_container_add (GTK_CONTAINER (frame11), scrolledwindow1);
	gtk_container_set_border_width (GTK_CONTAINER (scrolledwindow1), 6);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_SHADOW_ETCHED_IN);

	units_treeview_unit2 = gtk_tree_view_new ();
	gtk_widget_show (units_treeview_unit2);
	gtk_container_add (GTK_CONTAINER (scrolledwindow1), units_treeview_unit2);
	gtk_widget_set_size_request (units_treeview_unit2, 230, 200);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (units_treeview_unit2), FALSE);

	label92 = gtk_label_new (_("To Unit"));
	gtk_widget_show (label92);
	gtk_frame_set_label_widget (GTK_FRAME (frame11), label92);

	label90 = gtk_label_new (_("Unit Selector"));
	gtk_widget_show (label90);
	gtk_expander_set_label_widget (GTK_EXPANDER (units_expander), label90);

	dialog_action_area3 = GTK_DIALOG (units_dialog)->action_area;
	gtk_widget_show (dialog_action_area3);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area3), GTK_BUTTONBOX_END);

	units_button_close = gtk_button_new_from_stock ("gtk-close");
	gtk_widget_show (units_button_close);
	gtk_dialog_add_action_widget (GTK_DIALOG (units_dialog), units_button_close, GTK_RESPONSE_CLOSE);
	GTK_WIDGET_SET_FLAGS (units_button_close, GTK_CAN_DEFAULT);
	gtk_tooltips_set_tip (tooltips, units_button_close, _("Close this window"), NULL);

	units_button_convert = gtk_button_new ();
	gtk_widget_show (units_button_convert);
	gtk_dialog_add_action_widget (GTK_DIALOG (units_dialog), units_button_convert, 0);
	GTK_WIDGET_SET_FLAGS (units_button_convert, GTK_CAN_DEFAULT);
	gtk_tooltips_set_tip (tooltips, units_button_convert, _("Convert value and unit"), NULL);

	alignment17 = gtk_alignment_new (0.5, 0.5, 0, 0);
	gtk_widget_show (alignment17);
	gtk_container_add (GTK_CONTAINER (units_button_convert), alignment17);

	hbox104 = gtk_hbox_new (FALSE, 2);
	gtk_widget_show (hbox104);
	gtk_container_add (GTK_CONTAINER (alignment17), hbox104);

	image38 = gtk_image_new_from_stock ("gtk-convert", GTK_ICON_SIZE_BUTTON);
	gtk_widget_show (image38);
	gtk_box_pack_start (GTK_BOX (hbox104), image38, FALSE, FALSE, 0);

	label96 = gtk_label_new_with_mnemonic (_("Convert"));
	gtk_widget_show (label96);
	gtk_box_pack_start (GTK_BOX (hbox104), label96, FALSE, FALSE, 0);
	
	units_button_reverse = gtk_button_new ();
	gtk_widget_show (units_button_reverse);
	gtk_dialog_add_action_widget (GTK_DIALOG (units_dialog), units_button_reverse, 0);
	GTK_WIDGET_SET_FLAGS (units_button_reverse, GTK_CAN_DEFAULT);
	gtk_tooltips_set_tip (tooltips, units_button_reverse, _("Convert from the second (converted) value and unit to the first unit"), NULL);

	alignment18 = gtk_alignment_new (0.5, 0.5, 0, 0);
	gtk_widget_show (alignment18);
	gtk_container_add (GTK_CONTAINER (units_button_reverse), alignment18);

	hbox105 = gtk_hbox_new (FALSE, 2);
	gtk_widget_show (hbox105);
	gtk_container_add (GTK_CONTAINER (alignment18), hbox105);

	image39 = gtk_image_new_from_stock ("gtk-convert", GTK_ICON_SIZE_BUTTON);
	gtk_widget_show (image39);
	gtk_box_pack_start (GTK_BOX (hbox105), image39, FALSE, FALSE, 0);

	label97 = gtk_label_new_with_mnemonic (_("Reverse"));
	gtk_widget_show (label97);
	gtk_box_pack_start (GTK_BOX (hbox105), label97, FALSE, FALSE, 0);

	units_store = gtk_list_store_new(UNITS_N_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(units_store), UNITS_TITLE_COLUMN, string_sort_func, GINT_TO_POINTER(UNITS_TITLE_COLUMN), NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(units_store), UNITS_NAMES_COLUMN, string_sort_func, GINT_TO_POINTER(UNITS_NAMES_COLUMN), NULL);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(units_store), UNITS_TITLE_COLUMN, GTK_SORT_ASCENDING);
	
	gtk_tree_view_set_model(GTK_TREE_VIEW(units_treeview_unit1), GTK_TREE_MODEL(units_store));
	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_unit1));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
	GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(_("Name"), renderer, "text", UNITS_TITLE_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UNITS_TITLE_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(units_treeview_unit1), column);
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(_("Unit"), renderer, "text", UNITS_NAMES_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UNITS_NAMES_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(units_treeview_unit1), column);
	g_signal_connect((gpointer) selection, "changed", G_CALLBACK(on_units_treeview_unit1_selection_changed), NULL);
	gtk_tree_view_set_enable_search(GTK_TREE_VIEW(units_treeview_unit1), TRUE);
	
	gtk_tree_view_set_model(GTK_TREE_VIEW(units_treeview_unit2), GTK_TREE_MODEL(units_store));
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_unit2));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(_("Name"), renderer, "text", UNITS_TITLE_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UNITS_TITLE_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(units_treeview_unit2), column);
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(_("Unit"), renderer, "text", UNITS_NAMES_COLUMN, NULL);
	gtk_tree_view_column_set_sort_column_id(column, UNITS_NAMES_COLUMN);
	gtk_tree_view_append_column(GTK_TREE_VIEW(units_treeview_unit2), column);
	g_signal_connect((gpointer) selection, "changed", G_CALLBACK(on_units_treeview_unit2_selection_changed), NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(units_store), UNITS_TITLE_COLUMN, string_sort_func, GINT_TO_POINTER(UNITS_TITLE_COLUMN), NULL);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(units_store), UNITS_NAMES_COLUMN, string_sort_func, GINT_TO_POINTER(UNITS_NAMES_COLUMN), NULL);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(units_store), UNITS_TITLE_COLUMN, GTK_SORT_ASCENDING);
	gtk_tree_view_set_enable_search(GTK_TREE_VIEW(units_treeview_unit2), TRUE);

	categories_store = gtk_tree_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(GTK_TREE_VIEW(units_treeview_category), GTK_TREE_MODEL(categories_store));
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(units_treeview_category));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(_("Category"), renderer, "text", 0, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(units_treeview_category), column);
	g_signal_connect((gpointer) selection, "changed", G_CALLBACK(on_units_treeview_category_selection_changed), NULL);
	gtk_tree_view_column_set_sort_column_id(column, 0);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(categories_store), 0, string_sort_func, GINT_TO_POINTER(0), NULL);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(categories_store), 0, GTK_SORT_ASCENDING);
	
	g_signal_connect_swapped ((gpointer) units_dialog, "delete_event", G_CALLBACK (close_app), NULL);
	g_signal_connect_swapped ((gpointer) units_dialog, "destroy_event", G_CALLBACK (close_app), NULL);
	g_signal_connect ((gpointer) units_button_convert, "clicked", G_CALLBACK (on_units_button_convert_clicked), NULL);
	g_signal_connect ((gpointer) units_button_reverse, "clicked", G_CALLBACK (on_units_button_reverse_clicked), NULL);
	g_signal_connect ((gpointer) units_entry_to_val, "activate", G_CALLBACK (on_units_entry_to_val_activate), NULL);
	g_signal_connect ((gpointer) units_entry_from_val, "activate", G_CALLBACK (on_units_entry_from_val_activate), NULL);
	g_signal_connect ((gpointer) units_entry_to_val, "changed", G_CALLBACK (on_units_entry_to_val_changed), NULL);
	g_signal_connect ((gpointer) units_entry_from_val, "changed", G_CALLBACK (on_units_entry_from_val_changed), NULL);
	g_signal_connect ((gpointer) units_button_close, "clicked", G_CALLBACK (on_units_button_close_clicked), NULL);
	g_signal_connect ((gpointer) units_entry_from_unit, "activate", G_CALLBACK (on_units_entry_from_unit_activate), NULL);
	g_signal_connect ((gpointer) units_entry_to_unit, "activate", G_CALLBACK (on_units_entry_to_unit_activate), NULL);
	g_signal_connect ((gpointer) units_entry_to_unit, "changed", G_CALLBACK (on_units_entry_to_unit_changed), NULL);
	g_signal_connect ((gpointer) units_entry_from_unit, "changed", G_CALLBACK (on_units_entry_from_unit_changed), NULL);

	return units_dialog;

}


void load_preferences() {

	evalops.structuring = STRUCTURING_NONE;
	evalops.approximation = APPROXIMATION_APPROXIMATE;
	evalops.parse_options.angle_unit = ANGLE_UNIT_RADIANS;
	evalops.parse_options.dot_as_separator = CALCULATOR->default_dot_as_separator;
	printops.number_fraction_format = FRACTION_DECIMAL;
	printops.can_display_unicode_string_function = &can_display_unicode_string_function;
	printops.spacious = true;
	printops.allow_non_usable = false;
	printops.excessive_parenthesis = false;
	printops.use_unicode_signs = true;
	printops.negative_exponents = false;
	printops.lower_case_e = true;
	printops.min_exp = EXP_PRECISION;
	
	wget_args = "--quiet --tries=1";
	initial_from_unit = "";
	initial_to_unit = "";
	selected_unit_category = "";
	first_time = false;
	initial_width = 550;
	initial_height = 550;
	initially_expanded = true;
	gchar *gstr = g_build_filename(g_get_home_dir(), ".qalculate", NULL);
	DIR *dir = opendir(gstr);
	g_free(gstr);
	if(!dir) {
		first_qalculate_run = true;
		first_time = true;
		return;
	} else {
		first_qalculate_run = false;
		closedir(dir);
	}
	int version_numbers[] = {0, 9, 3};
	FILE *file = NULL;
	gchar *gstr2 = g_build_filename(g_get_home_dir(), ".qalculate", "qalculate-units.cfg", NULL);
	file = fopen(gstr2, "r");
	g_free(gstr2);
	if(file) {
		char line[10000];
		string stmp, svalue, svar;
		size_t i;
		int v;
		while(true) {
			if(fgets(line, 10000, file) == NULL)
				break;
			stmp = line;
			remove_blank_ends(stmp);
			if((i = stmp.find_first_of("=")) != string::npos) {
				svar = stmp.substr(0, i);
				remove_blank_ends(svar);
				svalue = stmp.substr(i + 1, stmp.length() - (i + 1));
				remove_blank_ends(svalue);
				v = s2i(svalue);
				if(svar == "version") {
					parse_qalculate_version(svalue, version_numbers);
				} else if(svar == "wget_args") {
					wget_args = svalue;
				} else if(svar == "dot_as_separator") {
					evalops.parse_options.dot_as_separator = v;
				} else if(svar == "from_unit") {
					initial_from_unit = svalue;
				} else if(svar == "to_unit") {
					initial_to_unit = svalue;
				} else if(svar == "selected_category") {
					selected_unit_category = svalue;
				} else if(svar == "show_selector") {
					initially_expanded = v;
				}
			}
		}
	} else {
		first_time = true;
	}
}


void save_preferences() {
	FILE *file = NULL;
	gchar *gstr = g_build_filename(g_get_home_dir(), ".qalculate", NULL);
	mkdir(gstr, S_IRWXU);
	g_free(gstr);
	gchar *gstr2 = g_build_filename(g_get_home_dir(), ".qalculate", "qalculate-units.cfg", NULL);
	file = fopen(gstr2, "w+");
	g_free(gstr2);
	fprintf(file, "\n[General]\n");
	fprintf(file, "version=%s\n", VERSION);	
	fprintf(file, "wget_args=%s\n", wget_args.c_str());
	fprintf(file, "dot_as_separator=%i\n", evalops.parse_options.dot_as_separator);
	fprintf(file, "from_unit=%s\n", gtk_entry_get_text(GTK_ENTRY(units_entry_from_unit)));
	fprintf(file, "to_unit=%s\n", gtk_entry_get_text(GTK_ENTRY(units_entry_to_unit)));
	fprintf(file, "selected_category=%s\n", selected_unit_category.c_str());
	fprintf(file, "show_selector=%i\n", gtk_expander_get_expanded(GTK_EXPANDER(units_expander)));
	fclose(file);
}


int main (int argc, char **argv) {

#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
#endif

	gtk_init(&argc, &argv);
	
	//create the almighty Calculator object
	new Calculator();

	load_preferences();
	
	create_units_dialog();
	while(gtk_events_pending()) gtk_main_iteration();
	
	//load global definitions
	if(!CALCULATOR->loadGlobalPrefixes() || !CALCULATOR->loadGlobalDefinitions("units.xml") || !CALCULATOR->loadGlobalFunctions() || !CALCULATOR->loadGlobalDataSets() || !CALCULATOR->loadGlobalVariables()) {
		g_print(_("Failed to load global definitions!\n"));
	}
	
	CALCULATOR->loadLocalDefinitions();

	g_timeout_add(100, on_display_errors_timeout, NULL);

	generate_units_tree_struct();
	update_units_tree();
	
	if(!initial_to_unit.empty()) convert_units();
	
	gtk_widget_grab_focus(units_entry_from_val);
	
	icon_pixbuf = gdk_pixbuf_new_from_xpm_data((const char**) icon_xpm);

	gtk_window_set_icon(GTK_WINDOW(units_dialog), icon_pixbuf);
	
	gtk_widget_show(units_dialog);
	
	gtk_main();
	
	return 0;
}
